(* These functions can be applied to the whole of an expression,
   in order to simplify cosines;
   the Head of the expression must be Plus in order to maximize
   the instances in which such simplifications will take place   *)

replacec2All::usage =
  "replacec2All[expr] acts on expression expr (whose Head must be Plus
  for maximum efficiency), replacing cosines by use of:
     c2(x)   = 1 - 2 s2(x/2)^2
     c2(x)^2 = 1 - s2(x)^2,
  provided that arguments of trigonometric functions remain integer
  multiples of momenta. Implicit sums over the directions of x are
  duly taken care of.
  "


replacec2All[a_Plus]:= replacec2All /@ a

replacec2All[c2[a_,b_]^i_ c_.]:= 
     replace[c2[a,b]^i c, {c2[a,b]^i, c2[a,b]^(i-2)}] -
      c2[a,b]^(i-2) s2[a,b]^2 c

replacec2All[c2[2 a_,b_] c_.]:=
     replace[c2[2a,b] c, {c2[2a,b],1}] - 2 canonical[s2[a,b]]^2 c

replacec2All[c2[4 a_,b_] c_.]:=
     replace[c2[4a,b] c, {c2[4a,b],1}] - 2 canonical[s2[2a,b]]^2 c

replacec2All[a_]:= a


replacec2q::usage =
  "replacec2q[expr] acts on expression expr (whose Head must be Plus
  for maximum efficiency), replacing cosines of (a multiple of) 
  external momentum q by use of:
     c2(a q)   = 1 - 2 a^2 s2(q)^2 + O(q^4)
     c2(a q)^2 = 1 -   a^2 s2(q)^2 + O(q^4),
  Implicit sums over the directions of x are duly taken care of.
  Terms of O(q^4) are ignored.
  "

replacec2q[a_Plus]:= replacec2q /@ a

replacec2q[c2[a_. q[j_],b_]^i_ c_.]:= 
     replace[c2[a q[j],b]^i c, {c2[a q[j],b]^i, c2[a q[j],b]^(i-2)}] -
      c2[a q[j],b]^(i-2) s2[q[j],b]^2 a^2 c

replacec2q[c2[a_. q[j_],b_] c_.]:=
     replace[c2[a q[j],b] c, {c2[a q[j],b],1}] - (a^2)/2 s2[q[j],b]^2 c

replacec2q[a_]:= a



replacec2qexact::usage =
  "replacec2qexact[expr] acts on expression expr (whose Head must be Plus
  for maximum efficiency), replacing cosines of (a multiple of) 
  external momentum q by use of the exact formulae:
     c2(a q)^2 = 1 -   s2(a q)^2
     c2(a q)   = 1 - 2 s2(a q/2)^2,
  Implicit sums over the directions of x are duly taken care of.
  "

replacec2qexact[a_Plus]:= replacec2qexact /@ a

replacec2qexact[c2[a_. q[j_],b_]^i_ c_.]:= 
     replace[c2[a q[j],b]^i c, {c2[a q[j],b]^i, c2[a q[j],b]^(i-2)}] -
      c2[a q[j],b]^(i-2) s2[a q[j],b]^2 c

replacec2qexact[c2[a_. q[j_],b_] c_.]:=
     replace[c2[a q[j],b] c, {c2[a q[j],b],1}] - 2 s2[a q[j]/2,b]^2 c

replacec2qexact[a_]:= a
